'From Squeak3.7beta of ''1 April 2004'' [latest update: #5923] on 20 September 2004 at 9:30:54 pm'!
"Change Set:		TimeZoneDatabaseV1-2-2
Date:			20 September 2004
Author:			David T. Lewis

TimeZoneDatabase integration with Squeak Kernel-Chronology.

Adds class TimeZoneProxy, which provides TimeZone protocol for the local
time zone transform in a TimeZoneDatabase. This permits methods such as
DateAndTime class>>now and Time class>>now to obtain the correct time zone
offset for any PointInTime, with compensation for daylight savings time
and other peculiarities.

Rather than relying on a static definition of the time zone, expressions
such as 'DateAndTime now' and 'Time now' will now perform a dynamic
lookup into the time zone database to determine the time zone offset. When
daylight savings time takes effect, the time zone database will automatically
compensate for the change.

Also, time zone offsets can now be expressed as Durations, as in:

  TimeZoneDatabase systemDatabase
    durationFrom: 'America/Detroit' to: 'Canada/Newfoundland'
    ==> -0:01:30:00

Here is how to load a time zone database and set the default time zone on a
Squeak image running on a typical Linux box:

1) Build a system database from files in the /usr/share/zoneinfo directory
(a typical location for a Unix/Linux system).

  TimeZoneDatabase buildSystemDatabaseForSqueak

2) Look for a suitable time zone to use. For example, if you live in Detroit, Michigan
in the USA, you can find several valid time zone entries for Detroit by evaluating
the following expression:

  TimeZoneDatabase systemDatabase grepFor: 'Detroit'

3) Select a time zone to use as your default. For example, if you have decided
to use 'America/Detroit', tell the database that this should be the default: 

  TimeZoneDatabase systemDatabase defaultLocation: 'America/Detroit'

4) Now tell Squeak to use this as its dynamic time zone setting.

  TimeZoneProxy
    setDynamicLocalTransform: TimeZoneDatabase systemDatabase defaultTimeZone
    abbreviation: 'EST'

"!

Object subclass: #PointInTime
	instanceVariableNames: 'absoluteTime '
	classVariableNames: 'NumericRepresentationSelector Platform '
	poolDictionaries: ''
	category: 'Time-UTC'!
Object subclass: #TimeZoneProxy
	instanceVariableNames: 'abbreviation name transform'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Time-TimeZones'!

!TimeZoneProxy commentStamp: '<historical>' prior: 0!
A TimeZone that obtains its information from an instance of TimeZoneRuleSet. Typically this is the the LocalTimeTransform for local time zone on this system, obtained from a TimeZoneDatabase.!


!PointInTime class methodsFor: 'initialize-release' stamp: 'dtl 9/19/2004 23:58'!
initialize
	"Set the selector to use to provide a floating point representation with
	adequate precision. Some systems make a distinction between Float and
	Double, and others simply use a double precision implementation of Float
	for all cases."

	"PointInTime initialize"

	Platform _ self platformSymbol.
	(Platform == #Squeak) ifTrue: [self numericRepresentationSelector: #asFloat].
	(Platform == #VisualWorks) ifTrue: [self numericRepresentationSelector: #asDouble].
	(Platform == #'Smalltalk/X') ifTrue: [PointInTime numericRepresentationSelector: #asFloat]! !

!PointInTime class methodsFor: 'private' stamp: 'dtl 9/19/2004 23:50'!
platform
	"Answer what version of Smalltalk we are using."

	"PointInTime platform"

	^ Platform
! !

!PointInTime class methodsFor: 'private' stamp: 'dtl 9/19/2004 23:56'!
platformSymbol
	"Answer what version of Smalltalk we are using."

	| versionString versionClass |
	(versionClass _ Smalltalk at: #SystemVersion ifAbsent: []) isNil
		ifTrue: [versionString _ Smalltalk version]
		ifFalse: [versionString _ versionClass current version].
	('Squeak*' match: versionString) ifTrue: [^ #Squeak].
	('VisualWorks*' match: versionString) ifTrue: [^ #VisualWorks].
	('*Smalltalk.st*' match: versionString) ifTrue: [^ #'Smalltalk/X'].
	^ #unknown! !


!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 9/13/2004 11:36'!
hoursFrom: name to: anotherTimeZoneName
	"Ask the database for the offset in hours between two time zones. The result
	may be an integer or a fraction."

	"TimeZoneDatabase systemDatabase hoursFrom: 'Europe/Amsterdam' to: 'America/Detroit'"
	"(TimeZoneDatabase systemDatabase hoursFrom: 'CET' to: 'Canada/Newfoundland') asFloat"

	^ (self offsetFrom: name to: anotherTimeZoneName) / 3600
! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 9/20/2004 19:32'!
offsetFrom: name to: anotherTimeZoneName
	"Ask the database for the current offset in seconds between two time zones."

	"TimeZoneDatabase systemDatabase offsetFrom: 'Europe/Amsterdam' to: 'America/Detroit'"

	| now |
	now _ PointInTime now.
	^ (self offsetFor: name at: now) - (self offsetFor: anotherTimeZoneName at: now)
! !

!TimeZoneDatabase methodsFor: '*kernel-Chronology' stamp: 'dtl 9/20/2004 19:32'!
durationFrom: name to: anotherTimeZoneName
	"Ask the database for the current offset between two time zones as a Duration."

	"TimeZoneDatabase systemDatabase durationFrom: 'Europe/Amsterdam' to: 'America/Detroit'"
	"TimeZoneDatabase systemDatabase durationFrom: 'CET' to: 'Canada/Newfoundland'"

	^ Duration seconds: (self offsetFrom: name to: anotherTimeZoneName)
! !


!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 9/13/2004 20:33'!
buildSystemDatabaseForSqueak
	"Build a full timezone database and install it as ThisSystemDatabase. Load the
	database from any tzfile files found in the default location for this system.
	Check first to determine if the time zone files exist, and issue a warning if
	the files are not available."

	"TimeZoneDatabase buildSystemDatabaseForSqueak"

	^ (FileDirectory on: self tzPrefixPath) entries isEmpty
		ifTrue: [self notify: 'no time zone files found in ', self tzPrefixPath]
		ifFalse: [self buildSystemDatabase]
! !


!TimeZoneProxy methodsFor: 'initialize-release' stamp: 'dtl 9/13/2004 12:43'!
initialize

	self name; abbreviation
! !

!TimeZoneProxy methodsFor: 'accessing' stamp: 'dtl 9/13/2004 12:47'!
abbreviation
	"Answer the value of abbreviation"

	^ abbreviation ifNil: [abbreviation _ self transform tzNameAbbreviations first asString]! !

!TimeZoneProxy methodsFor: 'accessing' stamp: 'dtl 9/13/2004 12:31'!
abbreviation: anObject
	"Set the value of abbreviation"

	abbreviation _ anObject! !

!TimeZoneProxy methodsFor: 'accessing' stamp: 'dtl 9/13/2004 12:42'!
name
	"Answer the value of name"

	^ name ifNil: [name _ self transform timeZoneName asString]
! !

!TimeZoneProxy methodsFor: 'accessing' stamp: 'dtl 9/13/2004 12:31'!
name: anObject
	"Set the value of name"

	name _ anObject! !

!TimeZoneProxy methodsFor: 'accessing' stamp: 'dtl 9/13/2004 12:31'!
transform
	"Answer the value of transform"

	^ transform! !

!TimeZoneProxy methodsFor: 'accessing' stamp: 'dtl 9/13/2004 12:31'!
transform: anObject
	"Set the value of transform"

	transform _ anObject! !

!TimeZoneProxy methodsFor: 'time zone offset' stamp: 'dtl 9/13/2004 13:11'!
offset

	^ Duration seconds: (self transform localOffsetSecondsAt: PointInTime now) negated! !

!TimeZoneProxy methodsFor: 'printing' stamp: 'dtl 9/13/2004 12:32'!
printOn: aStream

	super printOn: aStream.
	aStream
		nextPut: $(;
		nextPutAll: self abbreviation;
		nextPut: $).! !


!TimeZoneProxy class methodsFor: 'instance creation' stamp: 'dtl 9/13/2004 12:40'!
transform: aTimeZoneRuleSet

	"TimeZoneProxy transform: TimeZoneDatabase systemDatabase defaultTimeZone"

	^ (self basicNew transform: aTimeZoneRuleSet) initialize! !

!TimeZoneProxy class methodsFor: 'instance creation' stamp: 'dtl 9/13/2004 12:59'!
transform: aTimeZoneRuleSet abbreviation: abbreviation
	"A time zone rule set typically has more than one abbreviation. This allows a
	specific abbreviation to be used. The abbreviation must be one of the abbreviations
	listed in the rule set."

	"TimeZoneProxy transform: TimeZoneDatabase systemDatabase defaultTimeZone abbreviation: 'EST'"
	"TimeZoneProxy transform: TimeZoneDatabase systemDatabase defaultTimeZone abbreviation: 'XXX'"

	(aTimeZoneRuleSet tzNameAbbreviations includes: abbreviation asSymbol)
		ifTrue:
			[^ (self basicNew
				transform: aTimeZoneRuleSet;
				abbreviation: abbreviation asString) initialize]
		ifFalse:
			[self notify: abbreviation, ' not valid, must be one of ',
				aTimeZoneRuleSet tzNameAbbreviations asString]
! !

!TimeZoneProxy class methodsFor: 'set local time zone' stamp: 'dtl 9/20/2004 19:41'!
setDynamicLocalTransform: aTimeZoneRuleSet abbreviation: abbreviation
	"Set the local time zone to a proxy for the TimeZoneRuleSet that describes this
	time zone. The time zone offset value will be dynamically calculated for any
	point in time. This causes the time zone offset to be dynamically corrected for
	daylight savings time and other time zone anomolies.

	Using a time zone proxy causes methods such as DateAndTime class>>now and
	Time class>>now to run more slowly, which could be a concern in applications
	that do a very large number of system time queries."

	"TimeZoneProxy setDynamicLocalTransform: TimeZoneDatabase systemDatabase defaultTimeZone abbreviation: 'EST'"

	| proxy |
	proxy _ self transform: aTimeZoneRuleSet abbreviation: abbreviation.
	DateAndTime localTimeZone: proxy.
	^ proxy! !

!TimeZoneProxy class methodsFor: 'set local time zone' stamp: 'dtl 9/20/2004 19:40'!
setStaticLocalTransform: aTimeZoneRuleSet abbreviation: abbreviation
	"Set the local time zone to a TimeZone with an offset corresponding to the
	offset of the local time zone at the present time. Time zone offsets will not
	be dynamically calculated, and this TimeZone will not be valid across
	daylight savings time changes.
	
	Unless the speed of system time queries is a serious concern, the
	#setDynamicLocalTransform:abbreviation: method is preferred."

	"TimeZoneProxy setStaticLocalTransform: TimeZoneDatabase systemDatabase defaultTimeZone abbreviation: 'EST'"

	| proxy tz |
	proxy _ self transform: aTimeZoneRuleSet abbreviation: abbreviation.
	tz _ TimeZone
		offset: proxy offset
		name: proxy name
		abbreviation: proxy abbreviation.
	DateAndTime localTimeZone: tz.
	^ proxy! !


!TimeZoneProxy class reorganize!
('instance creation' transform: transform:abbreviation:)
('set local time zone' setDynamicLocalTransform:abbreviation: setStaticLocalTransform:abbreviation:)
!


!TimeZoneProxy reorganize!
('initialize-release' initialize)
('accessing' abbreviation abbreviation: name name: transform transform:)
('time zone offset' offset)
('printing' printOn:)
!


!TimeZoneDatabase reorganize!
('accessing' defaultLocation defaultLocation: indexSeparator indexSeparator: leapSecondRuleSet leapSecondRuleSet: timeZones)
('adding' indexAt:put: indexAt:withSeparator:put: loadDataFrom: loadDataFrom:prefixPath: loadDataFrom:prefixPath:pathSeparator:condenseDuplicates: updateLeapSecondRulesFrom:)
('querying' allTimeZones defaultTimeZone grepFor: hoursFrom:to: leapSecondsAt: leapSecondsFor:at: offsetFor:at: offsetFrom:to: timeZoneFor: timeZoneFor:pathSeparator: timeZoneNames)
('*kernel-Chronology' durationFrom:to:)
('normalizing' combineDuplicateTables: normalize setLeapSecondRulesFrom:)
('private' addZone:toHierarchy:withKeys: addZonesTo:from: isContainerNode: isTerminalLeaf: leapSecondsRuleFor: timeZoneIndexedBy:inHierarchy: tokensFor:withSeparator:)
!

PointInTime initialize!

!PointInTime class reorganize!
('initialize-release' initialize numericRepresentationSelector:)
('instance creation' date:time:timeZone: fromPosixSeconds: fromSmalltalkSeconds:inTimeZone: now)
('general inquiries' dateAndTimeFromSmalltalkSeconds: dateAndTimeNow dateFromSmalltalkSeconds:)
('testing' testFromSmalltalkSeconds testLocalSeconds)
('private' fromSmalltalkSeconds: platform platformSymbol posixOffset)
!

Object subclass: #PointInTime
	instanceVariableNames: 'absoluteTime'
	classVariableNames: 'NumericRepresentationSelector Platform'
	poolDictionaries: ''
	category: 'Time-UTC'!
